#!/usr/bin/perl

use strict;
use warnings;
use open ':encoding(utf8)'; #Equivalent to perl -CD,
binmode STDERR,":encoding(utf8)";
binmode STDOUT,":encoding(utf8)"; #Equivalent to -CS
binmode STDIN,":encoding(utf8)";

my $DOSTEP1=1;
my $DOSTEP2=1;
my $DOSTEP3=1;

my $NOPARSE=0;
my $POS_DEVIATIONS=0;

my $STANFORD_JAR="/home/users0/anders/storage/scratch/anders/stanford-parser-2012-07-09/stanford-parser.jar";

my $input=shift;
my $tmpFile=(scalar(@ARGV)?shift:"parse_trees.tmp");
my $tmpFile2=(scalar(@ARGV)?shift:"parse_trees.tmp.deps");

# just do it in two passes -- first pass, convert to CFG trees;
# second pass, feed it to the stanford thing and merge with the input
if($DOSTEP1){
    print STDERR "Writing CFG trees to $tmpFile\n";
    open CONLL,$input;
    open TMP,">",$tmpFile;
#my @lines;
    while(my $conll=<CONLL>){
	chomp $conll;
	next if($conll=~/^#/);
	if($conll=~/^\s*$/){
	    print TMP "\n";
	    next;
	}
	my @cols=split(/\s+/,$conll);
	if ($conll =~ m/^[^\#]\S+\s+\d+\s+\d+\s+(\S+)\s+(\S+)\s+(\S*)\*(\S*)\s+/) {
	    my $form=$1;
	    my $pos=$2;
	    my $label=$3;
	    my $end=$4;
	    if($label=~/NOPARSE$/){
		$NOPARSE++;
		$label=~s/NOPARSE$/VP/;
	    }
	    print TMP $label, " (", $pos, " ", $form, ")", $end, "\n";
	} else {
	    die("Failed to parse line: $conll");
	}
    }

    close CONLL;
    close TMP;
    print STDERR scalar(localtime()),"\n";
    print STDERR "Done extracting CFG trees\n";
    print STDERR "Number of NOPARSE: $NOPARSE\n";
}


if($DOSTEP2){
    print STDERR "\nRunning stanford converter. In: $tmpFile -- Out: $tmpFile2\n";
    my $pid=open DEPSERR,"java -cp $STANFORD_JAR edu.stanford.nlp.trees.EnglishGrammaticalStructure -treeFile $tmpFile -basic -conllx 2>&1 1>$tmpFile2 |";
    my $ERROR=0;
    while(<DEPSERR>){
	print STDERR $_;
	$ERROR=1;
    }
    if($ERROR){
	print STDERR "Had error converting -- aborting!\n";
	print STDERR "EXITING!\n";
	exit 1;
    }
    waitpid $pid,0;
    print STDERR scalar(localtime()),"\n";
    print STDERR "Done running stanford converter\n";
}


if($DOSTEP3){
    open DEPS,$tmpFile2;
    open CONLL,$input;
    while(my $conll=<CONLL>){
	chomp $conll;
	my @cols=split(/\s+/,$conll);
	if($conll=~/^#/){
	    print $conll,"\n";
	    next;
	}
	if($conll=~/^\s*$/){
	    print "\n";
	    #Also grab a line from deps
	    my $foo=<DEPS>;
#	    exit 0; #DEBUG
	    next;
	}
	#Otherwise we have a token. Match it and then print the merged line
	my $depLine=<DEPS>;
	chomp $depLine;
	my $outLine=matchAndMerge($conll,$depLine);
	print $outLine,"\n";
    }

    close DEPS;
    close CONLL;
    print STDERR scalar(localtime()),"\n";
    print STDERR "\nDone merging.\n";
    print STDERR "Got $POS_DEVIATIONS deviations in POS tags\n";
}


sub matchAndMerge {
    my $conll=shift;
    my $depLine=shift;
    my @c=split(/\s+/,$conll);
    my @d=split(/\s+/,$depLine);
    die("Not same surface form: '$c[3]' vs '$d[1]'") if($c[3] ne $d[1]);
    die("Not same POS: '$c[4]' vs '$d[3]'") if($c[4] ne $d[3] && !allowedPosDeviation($c[4],$d[3],$c[3]));
    my @out=@c[0..10];
    push @out,@d[6..7];
    push @out,@c[11..(scalar(@c)-1)];
    return join("\t",@out);
}

sub allowedPosDeviation {
    my $conllPOS=shift;
    my $stanfordPOS=shift;
    my $form=shift;
    if($conllPOS eq 'DT' && $stanfordPOS eq 'CC' && $form eq 'both'){
	$POS_DEVIATIONS++;
	return 1;
    } elsif($conllPOS eq 'PDT' && $stanfordPOS eq 'CC' && $form eq 'both'){
	$POS_DEVIATIONS++;
	return 1;
    }
    return 0;
}
