#!/usr/bin/perl

use strict;
use warnings;
use open ':encoding(utf8)'; #Equivalent to perl -CD,
binmode STDERR,":encoding(utf8)";
binmode STDOUT,":encoding(utf8)"; #Equivalent to -CS
binmode STDIN,":encoding(utf8)";

use Coref;


#my $IN_SENT_ONLY=1;
my $TAKE_POSSESIVES=1;
my $TAKE_REFLEXIVES=1;
my $TAKE_REST=1;

my $docName;
my @doc;
my $lineCount;
my $inDocCount=0;
my $docCount=0;

while(<>){
    $lineCount++;
    chomp;
    if($_=~/^#begin document/){
	$docName=$_;
	next;
    }
    $inDocCount++;
    if($_=~/^#end document/){
	doIt();
	$docCount++;
	@doc=();
#	die('end of doc');
	next;
    }
    my @cols=split(/\s+/,$_);
    push @doc,\@cols;
}


sub doIt {
#    my %cid2span=getSpanHash();
    my $hashref=Coref::getSpanHash(\@doc);
    my $senCount=0;
    for(my $lineNo=0;$lineNo<scalar(@doc);++$lineNo){
	if(scalar(@{$doc[$lineNo]})<2){
#	    print "BLANK: ",join ("\t",@{$doc[$lineNo]}),"\n";
	    $senCount++;
	    next;
	}
	my $tag=$doc[$lineNo]->[4];
	my $form=$doc[$lineNo]->[3];
	next if($tag!~/^PRP/);
	my $isPossessive=0;
	my $isReflexive=0;
	$isPossessive=1 if($tag eq 'PRP$');
	$isReflexive=1 if($form=~/sel(f|ves)$/);
	my $isOther=1 if($isPossessive==0 && $isReflexive==0);
	next if(!$TAKE_REFLEXIVES && $isReflexive);
	next if(!$TAKE_POSSESIVES && $isPossessive);
	next if(!$TAKE_REST && $isOther);

#	print $tag,"\n";
#	print $form,"\n\n";
	if($doc[$lineNo]->[-1]=~/\((\d+)\)/){ #It has an antecedent
	    my $CID=$1;
	    my @ants=getAnts($CID,$lineNo,$hashref,$senCount);
	    outputAnts($lineNo,$form,$senCount,\@ants);
	} else {
	    outputNullAnt($lineNo,$form,$senCount);
	}
    }
}

sub getAnts {
    my $CID=shift;
    my $lineNo=shift;
#    my %cid2span=%{(shift)};
    my $hashref=shift;
    #We require a mention to at least begin on, or before the same line
#    print "Looking up CID: ",$CID,"\n";
    my @cluster=@{$hashref->{$CID}};
    my @ants;
    foreach my $mention(@cluster){
	my ($beg,$end,$sentCount)=@{$mention};
	next if($beg==$lineNo && $end==$lineNo); #then its the same mention
	next if($beg>$lineNo); #then it begins after this mention
	my $ref=[$beg,$end,$sentCount];
	push @ants,$ref;
    }
    return @ants;
}

sub outputAnts {
    my $lineNo=shift;
    my $form=shift;
    my $senCount=shift;
    my @ants=@{(shift)};
    print "$docName\t$lineNo\t$form\t$senCount\tY";
    foreach my $a(@ants){
	my ($beg,$end,$sameSen)=@{($a)};
	if($beg==-1 || $end==-1){
	    print $form,"\n";
	    print $beg,"\n";
	    print $end,"\n";
	    die('problem here');
	}
	print "\t".join(",",$beg,$end,$sameSen);
    }
    print "\n";
}

sub outputNullAnt {
    my $lineNo=shift;
    my $form=shift;
    my $senCount=shift;
    print "$docName\t$lineNo\t$form\t$senCount\tN\n";
}
